# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib

from lollypop.logger import Logger
from lollypop.selectionlist import SelectionList
from lollypop.define import App, Type, SelectionListMask, StorageType, ViewType
from lollypop.shown import ShownLists
from lollypop.helper_gestures import GesturesHelper
from lollypop.view import View
from lollypop.utils import emit_signal, get_default_storage_type


class NoneView(View):
    """
        A view that do nothing
    """

    def __init__(self):
        View.__init__(self, StorageType.COLLECTION, ViewType.DEFAULT)

    @property
    def args(self):
        return None


class ListsContainer:
    """
        Selections lists management for main view
    """

    def __init__(self):
        """
            Init container
        """
        self.__left_list = None
        self.__right_list = None
        self._sidebar = SelectionList(SelectionListMask.SIDEBAR)
        self._sidebar.show()
        self._sidebar.listbox.connect("row-activated",
                                      self.__on_sidebar_activated)
        self._sidebar.connect("populated", self.__on_sidebar_populated)
        items = ShownLists.get(SelectionListMask.SIDEBAR)
        self._sidebar.populate(items)
        self._stack.connect("set-sidebar-id", self.__on_set_sidebar_id)
        self._stack.connect("set-selection-ids", self.__on_set_selection_ids)

    @property
    def sidebar(self):
        """
            Get first SelectionList
            @return SelectionList
        """
        return self._sidebar

    @property
    def left_list(self):
        """
            Get left selection list
            @return SelectionList
        """
        if self.__left_list is None:
            self.__left_list = SelectionList(SelectionListMask.FASTSCROLL)
            self.__left_list.listbox.connect("row-activated",
                                             self.__on_left_list_activated)
            self.__left_list.scrolled.set_size_request(300, -1)
        return self.__left_list

    @property
    def right_list(self):
        """
            Get right selection list
            @return SelectionList
        """
        def on_unmap(widget):
            """
                Hide right list on left list hidden
            """
            if not App().window.folded:
                self._hide_right_list()

        if self.__right_list is None:
            eventbox = Gtk.EventBox.new()
            eventbox.set_size_request(50, -1)
            eventbox.show()
            self.__right_list_grid = Gtk.Grid()
            style_context = self.__right_list_grid.get_style_context()
            style_context.add_class("left-gradient")
            style_context.add_class("opacity-transition-fast")
            self.__right_list = SelectionList(SelectionListMask.FASTSCROLL)
            self.__right_list.show()
            self.__right_list.scrolled.set_size_request(250, -1)
            self.__gesture = GesturesHelper(
                eventbox, primary_press_callback=self._hide_right_list)
            self.__right_list.listbox.connect("row-activated",
                                              self.__on_right_list_activated)
            self.__right_list_grid.add(eventbox)
            self.__right_list_grid.add(self.__right_list)
            self.__left_list.overlay.add_overlay(self.__right_list_grid)
            self.__left_list.connect("unmap", on_unmap)
            self.__left_list.connect("populated", self.__on_list_populated)
            self.__right_list.connect("populated", self.__on_list_populated)
        return self.__right_list

    @property
    def sidebar_id(self):
        """
            Get sidebar id for current state
            @return int
        """
        if self.right_list.selected_ids:
            ids = self.right_list.selected_ids
        elif self.left_list.selected_ids:
            ids = self.left_list.selected_ids
        else:
            ids = self.sidebar.selected_ids
        return ids[0] if ids else Type.NONE

##############
# PROTECTED  #
##############
    def _show_genres_list(self, selection_list):
        """
            Setup list for genres
            @param list as SelectionList
        """
        def load():
            genres = App().genres.get()
            return genres
        selection_list.set_mask(SelectionListMask.GENRES)
        App().task_helper.run(load, callback=(selection_list.populate,))

    def _show_artists_list(self, selection_list, genre_ids=[]):
        """
            Setup list for artists
            @param list as SelectionList
            @param genre_ids as [int]
        """
        def load():
            storage_type = get_default_storage_type()
            artists = App().artists.get(genre_ids, storage_type)
            return artists
        selection_list.set_mask(SelectionListMask.ARTISTS)
        App().task_helper.run(load, callback=(selection_list.populate,))

    def _show_right_list(self):
        """
            Show right list
        """
        if self.__right_list is not None:
            self.__right_list_grid.show()
            self.__right_list_grid.set_state_flags(Gtk.StateFlags.VISITED,
                                                   False)
            self.set_focused_view(self.right_list)

    def _hide_right_list(self, *ignore):
        """
            Hide right list
        """
        if self.__right_list is not None:
            self.__right_list_grid.unset_state_flags(Gtk.StateFlags.VISITED)
            GLib.timeout_add(200, self.__right_list_grid.hide)
            self.__right_list.clear()
            self.set_focused_view(self.left_list)

############
# PRIVATE  #
############
    def __on_sidebar_activated(self, listbox, row):
        """
            Update view based on selected object
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        Logger.debug("Container::__on_sidebar_activated()")
        self.sub_widget.set_visible_child(self.grid_view)
        view = None
        focus_set = False
        selected_id = self._sidebar.selected_id
        if selected_id is None:
            return
        # Update lists
        if selected_id in [Type.ARTISTS_LIST, Type.GENRES_LIST] and\
                self.type_ahead.get_reveal_child() and\
                self.left_list.get_visible():
            self.set_focused_view(self.left_list)
            focus_set = True
        elif selected_id == Type.ARTISTS_LIST:
            self._show_artists_list(self.left_list)
            self._hide_right_list()
            self.left_list.show()
            self.widget.set_visible_child(self.sub_widget)
            self.sub_widget.set_visible_child(self.left_list)
            self.set_focused_view(self.left_list)
            focus_set = True
        elif selected_id == Type.GENRES_LIST:
            self._show_genres_list(self.left_list)
            self._hide_right_list()
            self.left_list.show()
            self.widget.set_visible_child(self.sub_widget)
            self.sub_widget.set_visible_child(self.left_list)
            self.set_focused_view(self.left_list)
            focus_set = True
        else:
            self.left_list.hide()
            self.left_list.clear()
            if self.widget.get_folded():
                self.widget.set_visible_child(self.sub_widget)
                self.sub_widget.set_visible_child(self.grid_view)

        storage_type = get_default_storage_type()
        if selected_id in [Type.ARTISTS_LIST, Type.GENRES_LIST] and not\
                App().window.folded:
            view = NoneView()
            view.show()
        elif selected_id == Type.PLAYLISTS:
            view = self._get_view_playlists()
        elif selected_id == Type.LYRICS:
            view = self._get_view_lyrics()
        elif selected_id == Type.CURRENT:
            view = self.get_view_current()
        elif selected_id == Type.SEARCH:
            view = self.get_view_search()
        elif selected_id == Type.INFO:
            view = self._get_view_info()
        elif selected_id == Type.SUGGESTIONS:
            view = self._get_view_suggestions(storage_type)
        elif selected_id in [Type.POPULARS,
                             Type.LOVED,
                             Type.RECENTS,
                             Type.LITTLE,
                             Type.RANDOMS]:
            view = self._get_view_albums([selected_id], [], storage_type)
        elif selected_id == Type.WEB:
            view = self._get_view_albums([selected_id], [], StorageType.SAVED)
        elif selected_id == Type.YEARS:
            view = self._get_view_albums_decades(storage_type)
        elif selected_id == Type.GENRES:
            view = self._get_view_genres(storage_type)
        elif selected_id == Type.ARTISTS:
            view = self._get_view_artists_rounded(storage_type)
        elif selected_id == Type.ALL:
            view = self._get_view_albums([selected_id], [], storage_type)
        elif selected_id == Type.COMPILATIONS:
            view = self._get_view_albums([selected_id], [], storage_type)
        if view is not None and view not in self._stack.get_children():
            view.show()
            self._stack.add(view)
        if view is not None:
            self._stack.set_visible_child(view)
            if not focus_set:
                self.set_focused_view(view)
        emit_signal(self, "can-go-back-changed", self.can_go_back)

    def __on_sidebar_populated(self, selection_list):
        """
            @param selection_list as SelectionList
        """
        if App().settings.get_value("save-state"):
            self._stack.load_history()
            emit_signal(self, "can-go-back-changed", self.can_go_back)
        else:
            startup_id = App().settings.get_value("startup-id").get_int32()
            if startup_id == -1:
                selection_list.select_first()
            else:
                selection_list.select_ids([startup_id], True)

    def __on_left_list_activated(self, listbox, row):
        """
            Update view based on selected object
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        Logger.debug("Container::__on_left_list_activated()")
        selected_ids = self.left_list.selected_ids
        view = None
        storage_type = get_default_storage_type()
        if self.left_list.mask & SelectionListMask.GENRES:
            if not App().window.folded:
                view = self._get_view_albums(selected_ids, [], storage_type)
            self._show_artists_list(self.right_list, selected_ids)
            self._show_right_list()
        else:
            view = self._get_view_artists([], selected_ids, storage_type)
            self.set_focused_view(view)
            self.sub_widget.set_visible_child(self.grid_view)
        if view is not None:
            view.show()
            self._stack.add(view)
            self._stack.set_visible_child(view)

    def __on_right_list_activated(self, listbox, row):
        """
            Update view based on selected object
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        storage_type = get_default_storage_type()
        genre_ids = self.left_list.selected_ids
        artist_ids = self.right_list.selected_ids
        view = self._get_view_artists(genre_ids, artist_ids, storage_type)
        view.show()
        self._stack.add(view)
        self._stack.set_visible_child(view)
        self.set_focused_view(view)
        self.sub_widget.set_visible_child(self.grid_view)

    def __on_list_populated(self, selection_list):
        """
            Select pending ids
            @param selection_list as SelectionList
        """
        selection_list.select_pending_ids()

    def __on_set_sidebar_id(self, stack, sidebar_id):
        """
            Set sidebar id on container
            @param stack as ContainerStack
            @param sidebar_id as int
        """
        self._sidebar.select_ids([sidebar_id], False)

    def __on_set_selection_ids(self, stack, selection_ids):
        """
            Set sidebar id and left/right list ids
            @param stack as ContainerStack
            @param selection_ids as {"left": [int], "right": [int])
        """
        # Restore left list
        if selection_ids["left"]:
            if self.left_list.selected_ids != selection_ids["left"]:
                self.left_list.show()
                if self.left_list.count == 0:
                    self.left_list.set_selection_pending_ids(
                        selection_ids["left"])
                    if App().window.container.sidebar.selected_ids[0] ==\
                            Type.GENRES_LIST:
                        self._show_genres_list(self.left_list)
                    else:
                        self._show_artists_list(self.left_list)
                else:
                    self.left_list.select_ids(selection_ids["left"], False)
                    self.left_list.show()
        else:
            self.left_list.hide()
            self.left_list.clear()
        # Restore right list
        if selection_ids["right"]:
            if self.right_list.selected_ids != selection_ids["right"]:
                self.right_list.set_selection_pending_ids(
                                        selection_ids["right"])
                self._show_artists_list(self.right_list,
                                        selection_ids["left"])
                self._show_right_list()
        else:
            self._hide_right_list()



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Soup, Gio

import json
from hashlib import md5
from pickle import load, dump

from lollypop.helper_passwords import PasswordsHelper
from lollypop.logger import Logger
from lollypop.utils import get_network_available
from lollypop.define import LOLLYPOP_DATA_PATH, App
from lollypop.define import LASTFM_API_KEY, LASTFM_API_SECRET


class LastFMWebService:
    """
        Handle scrobbling to Last.fm and all authenticated API calls
    """

    def __init__(self, name):
        """
            Init service
            @param name as str
        """
        self.__name = name
        self.__queue = []
        if name == "LIBREFM":
            self.__uri = "https://libre.fm/2.0/"
        else:
            self.__uri = "https://ws.audioscrobbler.com/2.0/"
        self.start()

    def start(self):
        """
            Start web service (load save queue)
        """
        try:
            self.__cancellable = Gio.Cancellable()
            self.__queue = load(
                open(LOLLYPOP_DATA_PATH + "/%s_queue.bin" % self.__name, "rb"))
        except Exception as e:
            Logger.info("LastFMWebService::start(): %s", e)
            self.__queue = []

    def stop(self):
        """
            Stop current tasks and save queue to disk
            @return bool
        """
        self.__cancellable.cancel()
        try:
            with open(LOLLYPOP_DATA_PATH + "/%s_queue.bin" % self.__name,
                      "wb") as f:
                dump(list(self.__queue), f)
        except Exception as e:
            Logger.info("LastFMWebService::stop: %s", e)
        return True

    def listen(self, track, timestamp):
        """
            Submit a listen for a track (scrobble)
            @param track as Track
            @param timestamp as int
        """
        monitor = Gio.NetworkMonitor.get_default()
        if App().settings.get_value("disable-scrobbling") or\
                not get_network_available() or\
                monitor.get_network_metered():
            self.__queue.append((track, timestamp))
        elif track.id is not None and track.id >= 0:
            App().task_helper.run(self.__listen, track, timestamp)

    def playing_now(self, track):
        """
            Submit a playing now notification for a track
            @param track as Track
        """
        monitor = Gio.NetworkMonitor.get_default()
        if App().settings.get_value("disable-scrobbling") or\
                not get_network_available() or\
                monitor.get_network_metered():
            return
        if track.id is not None and track.id >= 0:
            App().task_helper.run(self.__playing_now, track)

    def love(self, artist, title):
        """
            Love track
            @param artist as string
            @param title as string
            @thread safe
        """
        App().task_helper.run(self.__love, artist, title, True)

    def unlove(self, artist, title):
        """
            Unlove track
            @param artist as string
            @param title as string
            @thread safe
        """
        App().task_helper.run(self.__love, artist, title, False)

    def set_loved(self, track, loved):
        """
            Add or remove track from loved playlist on Last.fm
            @param track as Track
            @param loved as bool
        """
        if loved == 1:
            self.love(",".join(track.artists), track.name)
        else:
            self.unlove(",".join(track.artists), track.name)

    def sync_loved_tracks(self):
        """
            Synced loved tracks from Last.fm
        """
        self.__passwords_helper = PasswordsHelper()
        self.__passwords_helper.get("LASTFM", self.__on_get_password)

#######################
# PRIVATE             #
#######################
    def __love(self, artist, title, status):
        """
            Love track
            @param artist as string
            @param title as string
            @param status as bool
        """
        try:
            token = App().ws_director.token_ws.get_token(
                self.__name, self.__cancellable)
            if token is None:
                return
            if status:
                args = self.__get_args_for_method("track.love")
            else:
                args = self.__get_args_for_method("track.unlove")
            args.append(("artist", artist))
            args.append(("track", title))
            args.append(("sk", token))
            api_sig = self.__get_sig_for_args(args)
            args.append(("api_sig", api_sig))
            post_data = {}
            for (name, value) in args:
                post_data[name] = value
            msg = Soup.form_request_new_from_hash("POST",
                                                  self.__uri,
                                                  post_data)
            msg.request_headers.append("Accept-Charset", "utf-8")
            data = App().task_helper.send_message_sync(msg, self.__cancellable)
            if data is not None:
                Logger.debug("%s: %s", self.__uri, data)
        except Exception as e:
            Logger.error("LastFMWebService::__love(): %s" % e)

    def __get_args_for_method(self, method):
        """
            Get arguments for method
            @param method as str
            @return [str]
        """
        args = [("method", method)]
        args.append(("api_key", LASTFM_API_KEY))
        return args

    def __get_sig_for_args(self, args):
        """
            Get API sig for method
            @param args as [str]
            @return str
        """
        args.sort()
        api_sig = ""
        for (name, value) in args:
            api_sig += "%s%s" % (name, value)
        api_sig += LASTFM_API_SECRET
        return md5(api_sig.encode("utf-8")).hexdigest()

    def __listen(self, track, timestamp):
        """
            Scrobble track
            @param track as Track
            @param timestamp as int
        """
        tracks = self.__queue + [(track, timestamp)]
        self.__queue = []
        try:
            for (track, timestamp) in tracks:
                token = App().ws_director.token_ws.get_token(
                    self.__name, self.__cancellable)
                if token is None:
                    return
                args = self.__get_args_for_method("track.scrobble")
                args.append(("artist", track.artists[0]))
                args.append(("albumArtist", track.album.artists[0]))
                args.append(("track", track.name))
                args.append(("album", track.album.name))
                if track.mbid and track.mbid.find(":") == -1:
                    args.append(("mbid", track.mbid))
                args.append(("timestamp", str(timestamp)))
                args.append(("sk", token))
                api_sig = self.__get_sig_for_args(args)
                args.append(("api_sig", api_sig))
                post_data = {}
                for (name, value) in args:
                    post_data[name] = value
                msg = Soup.form_request_new_from_hash("POST",
                                                      self.__uri,
                                                      post_data)
                msg.request_headers.append("Accept-Charset", "utf-8")
                data = App().task_helper.send_message_sync(msg,
                                                           self.__cancellable)
                if data is not None:
                    Logger.debug("%s: %s", self.__uri, data)
        except Exception as e:
            Logger.error("LastFMWebService::__listen(): %s" % e)

    def __playing_now(self, track):
        """
            Now playing track
            @param track as Track
        """
        try:
            token = App().ws_director.token_ws.get_token(
                self.__name, self.__cancellable)
            if token is None:
                return
            args = self.__get_args_for_method("track.updateNowPlaying")
            args.append(("artist", track.artists[0]))
            args.append(("albumArtist", track.album.artists[0]))
            args.append(("track", track.name))
            args.append(("album", track.album.name))
            if track.mbid and track.mbid.find(":") == -1:
                args.append(("mbid", track.mbid))
            args.append(("duration", str(track.duration // 1000)))
            args.append(("sk", token))
            api_sig = self.__get_sig_for_args(args)
            args.append(("api_sig", api_sig))
            post_data = {}
            for (name, value) in args:
                post_data[name] = value
            msg = Soup.form_request_new_from_hash("POST",
                                                  self.__uri,
                                                  post_data)
            msg.request_headers.append("Accept-Charset", "utf-8")
            data = App().task_helper.send_message_sync(msg, self.__cancellable)
            if data is not None:
                Logger.debug("%s: %s -> %s", self.__uri, data, post_data)
        except Exception as e:
            Logger.error("LastFMWebService::__playing_now(): %s" % e)

    def __populate_loved_tracks(self, user):
        """
            Populate loved tracks playlist
            @parma user as str
        """
        try:
            uri = "http://ws.audioscrobbler.com/2.0/"
            uri += "?method=user.getLovedTracks"
            uri += "&user=%s&api_key=%s&format=json&limit=1000" % (
                user, LASTFM_API_KEY)
            (status, data) = App().task_helper.load_uri_content_sync(uri, None)
            if status:
                content = json.loads(data.decode("utf-8"))
                for track in content["lovedtracks"]["track"]:
                    artist = track["artist"]["name"]
                    name = track["name"]
                    track_id = App().tracks.search_track(artist, name)
                    if track_id is None:
                        Logger.warning(
                            "LastFM: can't find %s, %s" % (
                                artist, name))
                    else:
                        App().tracks.set_loved(track_id, 1)
        except Exception as e:
            Logger.error("LastFM::__populate_loved_tracks: %s" % e)

    def __on_get_password(self, attributes, password, service):
        """
            Populate loved tracks
            @param attributes as {}
            @param password as str
            @param service as str
        """
        if attributes is not None:
            App().task_helper.run(
                self.__populate_loved_tracks, attributes["login"])
